from setuptools import setup, find_packages

setup(
    name="artypy",
    version="0.2.0",
    author="artypy team",
    author_email="ahmatovbulat@gmail.com",
    description="Lightweight library for artistic filters",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/artypy/arty",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
    install_requires=[
        "numpy",
        "matplotlib",
        "torch"
    ],
    extras_require={
        "core": []
    },
)



import numpy as np


def gauss(image: np.ndarray, kernel_size: int = 3, sigma: float = 1.0) -> np.ndarray:
    kernel = np.zeros((kernel_size, kernel_size))
    for x in range(kernel_size):
        for y in range(kernel_size):
            kernel[y, x] = np.exp(-((x - kernel_size // 2) ** 2 + (y - kernel_size // 2) ** 2) / (2 * sigma ** 2)) / (
                    2 * np.pi * sigma ** 2)
    kernel /= kernel.sum()

    kernel = np.array([
        [1, 2, 1],
        [2, 4, 2],
        [1, 2, 1]
    ]) / 16

    h, w, c = image.shape
    n = kernel_size
    new = image.copy()
    for y in range(n // 2, h - (n + 1) // 2):
        for x in range(n // 2, w - (n + 1) // 2):
            new[y, x] = [i.sum() for i in image[y - n // 2:y + (n + 1) // 2, x - n // 2:x + (n + 1) // 2].T * kernel]
    return new.astype('uint8')



import cv2
import math
import numpy as np


class Gradient:
    def __init__(self, image, type="sharr", type_smoothing="gaussian", radius=1, smoothing_iterations=1):
        """
        :param image: np.ndarray or cv2.UMat, grayscale or RGB image
        :param type: "sharr", "sobel", default is "sharr"
        :param type_smoothing: "gaussian", "bilateral", "none", default is "gaussian"
        :param radius: radius of the smoothing kernel, default is 1
        :param smoothing_iterations: number of smoothing iterations, default is 1
        """
        self.type = type
        self.type_smoothing = type_smoothing
        self.radius = radius

        # if image is np.ndarray, convert to cv2 image
        if isinstance(image, np.ndarray):
            if len(image.shape) == 2:
                gray_image = image
            elif len(image.shape) == 3:
                gray_image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            else:
                raise ValueError(f"Invalid image shape: {image.shape}, m ust be (h, w) or (h, w, 3)")
        # elif image is cv2 image, make it grayscale
        elif isinstance(image, cv2.UMat):
            if image.channels == 1:
                gray_image = image
            elif image.channels == 3:
                gray_image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            else:
                raise ValueError(f"Invalid image channels: {image.channels}, must be 1 or 3")
        else:
            raise ValueError(f"Invalid image type: {type(image)}, must be np.ndarray or cv2.UMat")

        if self.type == "sharr":
            self.grad_x = cv2.Scharr(gray_image, cv2.CV_32F, 1, 0)
            self.grad_y = cv2.Scharr(gray_image, cv2.CV_32F, 0, 1)
        elif self.type == "sobel":
            self.grad_x = cv2.Sobel(gray_image, cv2.CV_32F, 1, 0)
            self.grad_y = cv2.Sobel(gray_image, cv2.CV_32F, 0, 1)
        else:
            raise ValueError("Invalid gradient type: %s" % self.type)

        s = 2 * self.radius + 1
        for _ in range(smoothing_iterations):
            if self.type_smoothing == "gaussian":
                self.grad_x = cv2.GaussianBlur(self.grad_x, (s, s), 0)
                self.grad_y = cv2.GaussianBlur(self.grad_y, (s, s), 0)
            elif self.type_smoothing == "bilateral":
                self.grad_x = cv2.bilateralFilter(self.grad_x, s, 75, 75)
                self.grad_y = cv2.bilateralFilter(self.grad_y, s, 75, 75)
            elif self.type_smoothing == "none":
                pass
            else:
                raise ValueError("Invalid gradient smoothing type: %s" % self.type_smoothing)

        self.gradient = np.maximum(np.abs(self.grad_x), np.abs(self.grad_y))

    def angle(self, i, j):
        return np.arctan2(self.grad_y[i, j], self.grad_x[i, j])

    def strength(self, i, j):
        return np.sqrt(self.grad_x[i, j] ** 2 + self.grad_y[i, j] ** 2)


if __name__ == "__main__":
    import sys
    import matplotlib.pyplot as plt

    # example usage from root: python artypy[core]/edge/gradient.py _demo/images/img.png
    if len(sys.argv) != 2:
        print("Usage: python gradient.py <image>")
        sys.exit(1)

    image = cv2.imread(sys.argv[1])
    gradient_sharr = Gradient(image, type="sharr", type_smoothing="none", radius=20, smoothing_iterations=1)

    gradient_smoothed = Gradient(image, type="sharr", type_smoothing="gaussian", radius=20, smoothing_iterations=1)

    fig, ax = plt.subplots(2, 2)
    ax[0, 0].imshow(gradient_sharr.grad_x, cmap="gray")
    ax[0, 0].set_title("Gradient X (Sharr)")
    ax[0, 0].axis("off")

    ax[0, 1].imshow(gradient_sharr.grad_y, cmap="gray")
    ax[0, 1].set_title("Gradient Y (Sharr)")
    ax[0, 1].axis("off")

    ax[1, 0].imshow(gradient_smoothed.grad_x, cmap="gray")
    ax[1, 0].set_title("Gradient X (Smoothed)")
    ax[1, 0].axis("off")

    ax[1, 1].imshow(gradient_smoothed.grad_y, cmap="gray")
    ax[1, 1].set_title("Gradient Y (Smoothed)")
    ax[1, 1].axis("off")

    plt.show()


import os

# TODO: Add loading and saving functions
# TODO: Handle paths appropriately

class Preset:
    def __init__(self, img_path="../_demo/images/img.png", palette_size=10, stroke_scale=0, length_scale=1 / 3,
                 gradient_smoothing_radius=0, brush_type="circle", length="base", length_first_flag=True,
                 gradient_type="sharr", gradient_smoothing_type="gaussian", smoothing_iterations=1, grid_scale=3, grayscale=False, has_cardboard=False):
        self.img_path = img_path
        self.palette_size = palette_size
        self.stroke_scale = stroke_scale
        self.gradient_smoothing_radius = gradient_smoothing_radius
        self.brush_type = brush_type
        self.length = length
        self.length_first_flag = length_first_flag
        file_names = [int(file_name.split(".")[0]) for file_name in os.listdir(f"sbr/{str(self.brush_type)}")]

        if len(file_names) == 0:
            file_names = [-1]

        self.img_save_path = os.path.join(f"sbr/{str(self.brush_type)}/{str(max(file_names) + 1)}.jpg")
        self.preset_save_path = f"../configs/presets/sbr/{str(self.brush_type)}/{str(max(file_names) + 1)}.yaml"
        self.length_scale = length_scale
        self.gradient_type = gradient_type
        self.gradient_smoothing_type = gradient_smoothing_type
        self.smoothing_iterations = smoothing_iterations
        self.grid_scale = grid_scale
        self.grayscale = grayscale
        self.has_cardboard = has_cardboard


import core
import numpy as np
from core.edge.gradient import Gradient

def lines(image: np.ndarray, to_blur: bool = False, to_noise: bool = False) -> np.ndarray:
    if to_blur:
        image = core.filter.blur.gauss(image)

    if to_noise:
        image = core.filter.noise(image)

    gradient = Gradient(image)
    dx, dy = gradient.grad_x, gradient.grad_y
    edges = gradient.gradient
    angles = core.angle.gradient_truth(dx, dy)
    lines = core.filter.lines(image, angles, edges)

    return lines



import random
import scipy
import bisect

import cv2
import math
import numpy as np

from tqdm import tqdm

import matplotlib.pyplot as plt

from core.brush import Brush
from core.filter.color import generate_color_set, extend_color_set
from core.edge.gradient import Gradient


class RandomGrid:
    def __init__(self, height, width, scale):
        self.height = height
        self.width = width
        self.scale = scale

    def generate(self):
        radius = self.scale // 2

        grid = [
            ((y + random.randint(-radius, radius)) % self.height,
             (x + random.randint(-radius, radius)) % self.width)
            for y in range(0, self.height, self.scale)
            for x in range(0, self.width, self.scale)
        ]

        random.shuffle(grid)

        return grid


class ImagePainter:
    def __init__(self, image, preset):
        self.image = image
        self.preset = preset
        self.brush = Brush(self.preset.brush_type)
        self.stroke_scale = self._compute_stroke_scale()
        self.gradient_smoothing_radius = self._compute_smoothing_radius()
        self.color_set = None
        self.gradient = None
        self.k = 9
        self.result = None
        self.gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    def _compute_stroke_scale(self):
        return int(
            math.ceil(max(self.image.shape) / 1000)) if self.preset.stroke_scale == 0 else self.preset.stroke_scale

    def _compute_smoothing_radius(self):
        return int(round(
            max(self.image.shape) / 50)) if self.preset.gradient_smoothing_radius == 0 else self.preset.gradient_smoothing_radius

    def prepare_color_set(self):
        """Generate and extend the color set."""
        print("Computing color set...")
        image = self.gray_image if self.preset.grayscale else self.image
        max_img_size = 200
        self.color_set = generate_color_set(image, self.preset.palette_size, max_img_size)
        print("Extending color color set...")

        if not self.preset.grayscale:
            self.color_set = extend_color_set(self.color_set, [(0, 50, 0), (15, 30, 0), (-15, 30, 0)])

    def compute_gradient(self):
        """Compute the gradient of the image."""
        print("Computing gradient...")
        self.gradient = Gradient(self.gray_image, self.preset.gradient_type, self.preset.gradient_smoothing_type,
                                 self.gradient_smoothing_radius)

    def _color_probabilities(self, pixels):
        """Compute color probabilities for the given pixels."""
        distances = scipy.spatial.distance.cdist(pixels, self.color_set)
        inverted_distances = np.max(distances, axis=1, keepdims=True) - distances
        normalized_distances = inverted_distances / inverted_distances.sum(axis=1, keepdims=True)
        scaled_distances = np.exp(self.k * len(self.color_set) * normalized_distances)
        probabilities = scaled_distances / scaled_distances.sum(axis=1, keepdims=True)

        return np.cumulative_sum(probabilities, axis=1, dtype=np.float32)

    def _get_color(self, probabilities):
        """Select a color from the set based on probabilities."""
        r = random.uniform(0, 1)
        i = bisect.bisect_left(probabilities, r)

        return self.color_set[i] if i < len(self.color_set) else self.color_set[-1]

    def paint(self):
        """Perform the painting operation."""
        print("Painting image...")

        if self.preset.has_cardboard:
            result = cv2.medianBlur(self.image, 11) if not self.preset.grayscale else cv2.medianBlur(self.gray_image, 11)
        else:
            result = np.full_like(self.image, 255, dtype=np.uint8)

        grid = RandomGrid(self.image.shape[0], self.image.shape[1], scale=self.preset.grid_scale).generate()

        batch_size = 10000

        for h in tqdm(range(0, len(grid), batch_size)):
            batch = grid[h:min(h + batch_size, len(grid))]
            pixels = np.array([self.image[y, x] for y, x in batch])
            color_probabilities = self._color_probabilities(pixels)

            for i, (y, x) in enumerate(batch):
                color = self._get_color(color_probabilities[i])
                angle = math.degrees(self.gradient.angle(y, x)) + 90

                if self.preset.length == "base":
                    length = max(int(round(self.stroke_scale + self.stroke_scale * math.sqrt(
                        self.gradient.strength(y, x))) * self.preset.length_scale), 1)
                elif self.preset.length == "inverse":
                    length = max(
                        1,
                        int(1 / round(self.stroke_scale + self.stroke_scale * math.sqrt(
                            self.gradient.strength(y, x))) ** 1.9 * 10 * self.preset.length_scale)
                    )
                else:
                    raise ValueError(f"Invalid length function: {self.preset.length}")

                self.brush.apply(result, (x, y), length, color, self.stroke_scale, angle, self.preset.length_first_flag)

        self.result = result

        return result

    # function to show the result
    def show_result(self):
        plt.figure(figsize=(10, 10))
        plt.subplot(1, 1, 1)
        if self.preset.grayscale:
            plt.imshow(self.result)
        else:
            plt.imshow(cv2.cvtColor(self.result, cv2.COLOR_BGR2RGB))
        plt.show()



from torch import nn
from ._generate import get_batch
import torch

def model(path, kernel_size, device):
    try:
        model = torch.load(path)
    except FileNotFoundError:
        model = nn.Sequential(
            nn.Conv2d(3, kernel_size ** 2, kernel_size),
            nn.ReLU(),
            nn.Conv2d(kernel_size ** 2, 3, 1),
            nn.Sigmoid(),
        ).to(device)

        torch.save(model, path)

    return model


def train(model, kernel_size, device, path, max_epochs=-1):
    optim = torch.optim.Adam(model.parameters())
    loss_fn = nn.MSELoss()

    best = 1
    remaining = 1000

    epoch = 0
    while remaining:
        epoch += 1
        if max_epochs != -1 and epoch > max_epochs:
            torch.save(model, path)
            break

        input, output = get_batch(kernel_size)
        input = torch.tensor(input).to(device).permute(0, 3, 2, 1).float()
        output = torch.tensor(output).to(device).float()

        out = model(input)
        loss = loss_fn(out, output[:, :, None, None])

        remaining -= 1
        if loss < best:
            remaining += 10
            best = loss
            torch.save(model, path)

        optim.zero_grad()
        loss.backward()
        optim.step()

    return model



import cv2
import numpy as np
import math
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt


def variate_color(img: np.ndarray, hue_shift: int = 0, saturation_shift: int = 0,
                  luminosity_shift: int = 0) -> np.ndarray:
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV_FULL)
    shifts = [hue_shift, saturation_shift, luminosity_shift]

    for i, shift in enumerate(shifts):
        if shift != 0:
            hsv[:, :, i] = np.clip((hsv[:, :, i].astype(np.int16) + shift), 0, 255).astype(np.uint8)

    return cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR_FULL)


def generate_color_set(img: np.ndarray, n: int, max_img_size: int = 200) -> np.ndarray:
    ratio = min(1.0, float(max_img_size) / img.shape[1], float(max_img_size) / img.shape[0])
    resized_img = cv2.resize(img, (int(img.shape[1] * ratio), int(img.shape[0] * ratio),), interpolation=cv2.INTER_AREA)
    clt = KMeans(n_clusters=n, n_init=10)
    clt.fit(resized_img.reshape(-1, 3))

    return clt.cluster_centers_


def extend_color_set(colors: np.ndarray, extensions: list[tuple[int, int, int]]) -> np.ndarray:
    reshaped_colors = colors.reshape((1, len(colors), 3)).astype(np.uint8)
    extended_colors = [variate_color(reshaped_colors, *x).reshape((-1, 3)) for x in extensions]
    all_colors = np.vstack([colors.reshape((-1, 3))] + extended_colors)

    return all_colors


def visualize_color_set(colors: np.ndarray, cell_size: int = 80) -> np.ndarray:
    n = len(colors)
    cols = 10
    rows = math.ceil(n / cols)
    res = np.full((rows * cell_size, cols * cell_size, 3), 255, dtype=np.uint8)

    for i, color in enumerate(colors):
        y, x = divmod(i, cols)
        y0, y1 = y * cell_size, (y + 1) * cell_size
        x0, x1 = x * cell_size, (x + 1) * cell_size
        res[y0:y1, x0:x1] = color

    plt.figure(figsize=(10, 4))
    plt.imshow(cv2.cvtColor(res, cv2.COLOR_BGR2RGB))
    plt.show()

    return res


if __name__ == "__main__":
    img = cv2.imread("../../_demo/images/img.png")
    colors = generate_color_set(img, 10)
    extended_colors = extend_color_set(colors, [(0, 50, 0), (15, 30, 0), (-15, 30, 0)])
    visualize_color_set(extended_colors)

    print("Done")



import numpy as np
import math
import cv2


class Brush:
    def __init__(self, brush_type="circle", texture_image=None):
        self.brush_type = brush_type
        self.texture_image = texture_image

    def apply(self, res, position, size, color, stroke_scale=1, angle=0, length_first_flag=True):
        brush_methods = {
            "square": self._square,
            "rectangle": self._rectangle,
            "star": self._star,
            "circle": self._circle,
            "ellipse": self._ellipse,
            "line": self._line,
            "texture": self._texture,
        }

        if self.brush_type in brush_methods:
            brush_methods[self.brush_type](res, position, size, color, stroke_scale, angle, length_first_flag)
        else:
            raise ValueError(f"Invalid brush type: {self.brush_type}")

    def _square(self, res, position, size, color, *_):
        x, y = position
        half_size = size // 2
        cv2.rectangle(res, (x - half_size, y - half_size), (x + half_size, y + half_size), color, -1)

    def _rectangle(self, res, position, size, color, stroke_scale, angle, *_):
        box = cv2.boxPoints(((position[0], position[1]), (size, stroke_scale), angle))
        cv2.fillPoly(res, [box.astype(np.int32)], color)

    def _star(self, res, position, size, color, *_):
        x, y = position
        points = []
        for i in range(5):
            angle = i * 72  # 360 / 5 points for a star
            dx = int(size * math.cos(math.radians(angle)))
            dy = int(size * math.sin(math.radians(angle)))
            points.append((x + dx, y + dy))
        points = np.array(points, np.int32)
        cv2.fillPoly(res, [points], color)

    def _circle(self, res, position, size, color, *_):
        cv2.circle(res, position, size, color, -1, cv2.LINE_AA)

    def _ellipse(self, res, position, size, color, stroke_scale, angle, *_):
        cv2.ellipse(res, position, (size, stroke_scale), angle, 0, 360, color, -1, cv2.LINE_AA)

    def _line(self, res, position, size, color, stroke_scale, angle, length_first_flag):
        x, y = position

        if length_first_flag:
            start_point = (int(x + size * math.cos(math.radians(angle))), int(y + size * math.sin(math.radians(angle))))
            end_point = (int(x - size * math.cos(math.radians(angle))), int(y - size * math.sin(math.radians(angle))))
        else:
            start_point = (
                int(x + stroke_scale * math.cos(math.radians(angle))),
                int(y + stroke_scale * math.sin(math.radians(angle))),
            )
            end_point = (
                int(x - stroke_scale * math.cos(math.radians(angle))),
                int(y - stroke_scale * math.sin(math.radians(angle))),
            )

        cv2.line(res, start_point, end_point, color, stroke_scale, cv2.LINE_AA)

    def _texture(self, res, position, size, color, *_):
        if self.texture_image is None:
            raise ValueError("Texture image not provided for texture brush.")

        x, y = position
        half_size = size // 2
        texture = cv2.resize(self.texture_image, (size, size))
        roi = res[y - half_size: y + half_size, x - half_size: x + half_size]

        if roi.shape == texture.shape:
            np.copyto(roi, texture)


if __name__ == "__main__":
    # use brushes on the canvas
    canvas = np.zeros((512, 512, 3), np.uint8)

    # split into 6 different spaces on the canva
    positions = [(128, 128), (384, 128), (128, 384), (384, 384), (256, 128), (256, 384)]


    brush1 = Brush(brush_type="circle")
    brush1.apply(canvas, positions[0], 50, (255, 0, 0))

    brush2 = Brush(brush_type="rectangle")
    brush2.apply(canvas, positions[1], 50, (0, 255, 0), 10, 45)

    brush3 = Brush(brush_type="star")
    brush3.apply(canvas, positions[2], 50, (0, 0, 255))

    brush4 = Brush(brush_type="ellipse")
    brush4.apply(canvas, positions[3], 50, (255, 255, 0), 10, 45)

    brush5 = Brush(brush_type="line")
    brush5.apply(canvas, positions[4], 50, (255, 0, 255), 10, 45, True)

    brush6 = Brush(brush_type="line")
    brush6.apply(canvas, positions[5], 50, (0, 255, 255), 10, 45, False)

    # show
    cv2.imshow("Canvas", canvas)
    cv2.waitKey(0)







import numpy as np
import pygame
from math import radians, cos, sin


def lines(image: np.ndarray, angles: np.ndarray, edges: np.ndarray) -> np.ndarray:
    priority = [(edges[y, x], (x, y)) for y in range(len(edges)) for x in range(len(edges[y]))]
    h, w, c = image.shape
    new = pygame.Surface((w, h))
    for _, (x, y) in priority[::-1]:
        if new.get_at((x, y)) != (0, 0, 0, 255): continue
        angle = radians(angles[y, x])
        dy, dx = cos(angle) * 5, sin(angle) * 5
        pygame.draw.line(new, image[y, x], (x - dx, y - dy), (x + dx, y + dy), 3)
    return pygame.surfarray.pixels3d(new).transpose(1, 0, 2)



import numpy as np
from ._points import points
from ..edge import gradient
import pygame


def strokes(image: np.ndarray, angles: np.ndarray, mx_length=10) -> np.ndarray:
    pts = points(image, 7, 1, 0)
    grad = gradient(image) / 765
    image = 1 - image.mean(axis=2) / 255
    h, w = pts.shape

    surf = pygame.Surface((w, h))
    surf.fill((255, 255, 255))

    angles[grad < 0.3] = np.random.random(angles.shape)[grad < 0.3] * 180
    angles = np.radians(angles)
    X, Y = np.cos(angles), np.sin(angles)
    for y in range(h):
        for x in range(w):
            if pts[y, x]: continue
            ln = image[y, x] * grad[y, x] * mx_length / 2
            dx, dy = X[y, x], Y[y, x]
            pygame.draw.aaline(surf, (50, 50, 50), (x + ln * dy, y + ln * dx), (x - ln * dy, y - ln * dx))
    return pygame.surfarray.pixels3d(surf).transpose(1, 0, 2)



import numpy as np
from math import degrees, atan2


def gradient_truth(gradientX: np.ndarray, gradientY: np.ndarray, kernel_size: int = 5) -> np.ndarray:
    h, w = gradientX.shape
    res = np.zeros_like(gradientX)
    n = kernel_size
    for y in range(h):
        for x in range(w):
            dx = gradientX[max(0, y - n // 2):y + (n + 1) // 2, max(0, x - n // 2):x + (n + 1) // 2 - 1].sum()
            dy = gradientY[max(0, y - n // 2):y + (n + 1) // 2 - 1, max(0, x - n // 2):x + (n + 1) // 2].sum()
            res[y, x] = -degrees(atan2(dy, dx)) % 180
    return res



from random import random
import numpy as np


def angle2line(angle):
    angle = np.radians(angle)
    x1, y1 = np.cos(angle), np.sin(angle)
    x2, y2 = -x1, -y1
    b = x2 - x1
    a = y1 - y2

    return a, b, random() - 0.5


def angle2array(angle):
    angle %= 180
    i1, i2, i3 = angle < 60, (60 <= angle) & (angle < 120), (120 <= angle) & (angle <= 180)
    array = np.zeros((angle.shape[0], 3))
    array[i1] = np.array([1 - angle / 60, angle / 60, angle * 0]).T[i1]

    angle -= 60
    array[i2] = np.array([angle * 0, 1 - angle / 60, angle / 60]).T[i2]

    angle -= 60
    array[i3] = np.array([angle / 60, angle * 0, 1 - angle / 60]).T[i3]

    return array


def array2angle(array):
    ind = array.argmin(1)
    array[np.arange(ind.size), ind] = 0
    array /= array.sum(1)[:, None]
    add = (ind + 1) % 3
    ind = (ind - 1) % 3
    angles = 60 * (add + array[np.arange(ind.size), ind])

    return angles


def draw(angles, colors, size):
    X, Y = (size - 1) / 2, (size - 1) / 2
    lines = sorted([angle2line(angle) for angle in angles], key=lambda line: line[-1])
    y, x = np.mgrid[0:size, 0:size]
    y, x = y - Y, x - X
    kernel = np.ones((size, size, 3)) * colors[0]

    for (a, b, c), color in zip(lines, colors[1:]):
        kernel[a * y + b * x + c * size < 0] = color

    kernel = np.clip(kernel + (np.random.random(kernel.shape) - 0.5) * 0.2, 0, 1)

    return kernel


def get_batch(kernel_size, batch_size=32):
    input, output = [], []

    for _ in range(batch_size):
        angle = random() * 180
        angles = [angle + (random() - 0.5) * 30]
        colors = [np.random.random(3), np.random.random(3)]

        while random() < 0.25:
            colors.append(np.random.random(3))
            angles.append(angle + (random() - 0.5) * 30)

        input.append(draw(angles, colors, kernel_size))
        output.append(sum(angles) / len(angles))

    input = np.array(input)
    output = np.array(output)
    output = angle2array(output)

    return input, output



import numpy as np
import torch

from core.angle import cnn



class CNN:
    def __init__(self, kernel_size: int = 5, path: str = ''):
        self.kernel_size = kernel_size
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.path = path
        if self.path == '':
            self.path = f'{kernel_size}.pt'

        self.model = cnn.model(self.path, kernel_size, device=self.device)

    def fit(self, max_epochs: int = -1):
        self.model = cnn.train(self.model, self.kernel_size, self.device, self.path, max_epochs)

    def predict(self, image: np.ndarray) -> np.ndarray:
        h, w, c = image.shape
        s, e = (self.kernel_size - 1) // 2, self.kernel_size // 2
        img = np.zeros((h + s + e, w + s + e, c))
        img[s:-e, s:-e] = image

        img = torch.tensor(img).to(self.device).float().T / 255

        angle = self.model(img).T
        shape = angle.shape[:-1]
        angle = angle.reshape(-1, 3).cpu().detach().numpy()
        angle = cnn.array2angle(angle).reshape(shape)

        return angle



import numpy as np
from ..edge import gradient


def points(image: np.ndarray, max_distance: int = 7, gradient_scaler: float = 0,
           darkness_scaler: float = 0) -> np.ndarray:
    gradient_scaler = np.exp(gradient_scaler)
    darkness_scaler = np.exp(darkness_scaler)

    edges = _gradient(image) / 765
    priority = [(edges[y, x], (x, y)) for y in range(len(edges)) for x in range(len(edges[y]))]
    priority.sort(reverse=True)

    image = image.mean(axis=2) / 255

    h, w = image.shape
    pts = np.zeros((h + max_distance * 2, w + max_distance * 2))

    y, x = np.mgrid[-max_distance:max_distance + 1, -max_distance:max_distance + 1]
    dist = (x ** 2 + y ** 2) ** 0.5

    for edge, (x, y) in priority:
        r = max(1, max_distance * (1 - edge) ** gradient_scaler * image[y, x] ** darkness_scaler)

        area = pts[y:y + max_distance * 2 + 1, x:x + max_distance * 2 + 1] * dist
        area = area[np.nonzero(area)]

        if len(area) == 0 or area.min() > r:
            pts[y + max_distance, x + max_distance] = 1
    pts = pts[max_distance:-max_distance, max_distance:-max_distance]
    return 255 - pts * 255
